#include "bsp.h"
#include "tty.h"

typedef struct {
  word samples[2048][4];
  word sample_counter;
  word trigger_counter;
  word trigger_position;
  word status;
} LogicAnalyzer;

/* status bits */
#define LA_RUN		0x0001
#define LA_TRIGGER	0x0002
#define LA_RESET	0x0004

static word seed;

static word
random ()
{
  word b0 = (seed >> 3) ^ (seed >> 12) ^ (seed >> 14) ^ (seed >> 15);
  seed = (seed << 1) | (b0 & 1);
  return seed;
}

typedef union {
  byte b[4];
  long l;
  void *p;
} LongPtr;

static inline __attribute__((always_inline)) void * 
long2ptr(long l)
{
  LongPtr lp;
  lp.l = l;
  p10.b = (*(word *)(&lp.b[1])-0x4000) >> 7;
  return lp.p;
}
extern inline void *
byte2ptr(byte h, byte m, byte l)
{
  LongPtr lp;
  lp.b[0] = l;
  lp.b[1] = m;
  lp.b[2] = h;
  lp.b[3] = 0;
  p10.b = lp.b[2];
  return lp.p;
}

#define F_IO_0	p9.b0
#define F_IO_1	p9.b1
#define F_IO_2	p9.b2
#define F_IO_3	p9.b3

extern void dllexit();

#define LA_ADDR 0x00100000L
#define la ((volatile LogicAnalyzer *)(LA_ADDR))

#define RAM_ADDR 0x00400000L
volatile word *ram = (word *)(RAM_ADDR);
#define RAMSIZE (0x800000L)

#define LAS(x)	  la->status = (x); la->status = 0
#define RUN()	  LAS(LA_RESET); LAS(LA_RUN);
#define TRIGGER() LAS(LA_TRIGGER);

static int ndumps;

la_dump()
{
  word d0, d1, d2, i, t;

  if (++ndumps > 10)
    return;

  t = la->trigger_position;

  d2 = ~0;
  tty_rawc(128);
  for (i=t; i != t+1024; i++)
    {
      word addr = i & 0x3ff;
      d0 = la->samples[addr][0];
      d1 = la->samples[addr][1];
      d2 = la->samples[addr][2];
	  
      tty_rawc (i-t);
      tty_rawc ((i-t) >> 8);
      tty_rawc (d0);
      tty_rawc (d0 >> 8);
      tty_rawc (d1);
      tty_rawc (d1 >> 8);
      tty_rawc (d2);
    }
  tty_rawc(255);
  tty_rawc(255);
}

static inline void
wb (long a, word d)
{
  word *r = (word *)long2ptr (a + RAM_ADDR);
  *r = d;
}

static inline word
rb (long a)
{
  word x, y;
  volatile word *r = (word *)long2ptr (a + RAM_ADDR);
  x = *r;
  y = *r;
  la->status = LA_TRIGGER;
  return x;
}

#define RBE(a, d, w)				\
  d = rb(a);					\
  TRIGGER();					\
  if (d != w) {					\
    word d2 = rb(a);				\
    FAIL (a, d, d2, w);				\
    la_dump();					\
  }						\
  RUN();
  

int nfail;
int totalfail;

int
fail (long a, word d, word d2, word e)
{
  nfail ++;
  totalfail ++;
  if (nfail == 1)
    cprintf("\n");
  if (nfail < 20)
    {
      cprintf("\033[035mFAIL\033[0m: %X\tgot \033[31m%w %W\033[0m\n", a, d, d);
      cprintf("\t\texp \033[32m%w %W\033[0m\n", e, e);
      if (d2 == e)
	cprintf("\t\td2  \033[32m%w %W\033[0m\n", d2, d2);
      else
	cprintf("\t\td2  \033[31m%w %W\033[0m\n", d2, d2);
    }
  else if (nfail == 20)
    cprintf("... more failures happened\n");
  else if (nfail > 50)
    return 1;
  return 0;
}
#define FAIL(a,d,d2,e) if (fail(a,d,d2,e)) return

void
report ()
{
  if (nfail == 0)
    cprintf("\033[32mPASS");
  else if (nfail == 1)
    cprintf("\033[33m1 FAILURE");
  else if (nfail >= 50)
    cprintf("\033[31mMANY FAILURES");
  else
    cprintf("\033[31m%d FAILURES", nfail);
  if (nfail)
    cprintf(" (%d so far)", totalfail);
  cprintf("\033[0m\n");
}

static void
constant_test (word w)
{
  long a;
  word d;

  nfail = 0;
  cprintf("all %w...", w);
  for (a=0; a<RAMSIZE; a+=2)
    {
      wb(a, w);
    }
  for (a=0; a<RAMSIZE; a+=2)
    {
      RBE(a, d, w);
    }
  if (nfail)
    for (a=0; a<RAMSIZE; a+=2)
      {
	wb(a, w);
	RBE(a, d, w);
      }
  report();
}

static void
zeros_test ()
{
  constant_test(0x0000);
}

static void
ones_test ()
{
  constant_test(0xffff);
}

static void
aa55_test ()
{
  constant_test(0xaaaa);
  constant_test(0x5555);
}

static void
msb_test ()
{
  long a;
  word d, w;

  nfail = 0;
  cprintf("MSB...", w);
  for (a=0; a<RAMSIZE; a+=2)
    {
      LongPtr lp;
      lp.l = a;
      w = (lp.b[2] << 8) | lp.b[1];
      wb(a, w);
    }
  for (a=0; a<RAMSIZE; a+=2)
    {
      LongPtr lp;
      lp.l = a;
      w = (lp.b[2] << 8) | lp.b[1];
      RBE(a, d, w)
    }
  report();
}

static void
lsb_test ()
{
  long a;
  word d, w;

  nfail = 0;
  cprintf("LSB...", w);
  for (a=0; a<RAMSIZE; a+=2)
    {
      w = (word)a;
      wb(a, w);
    }
  for (a=0; a<RAMSIZE; a+=2)
    {
      w = (word)a;
      RBE(a, d, w)
    }
  report();
}

static void
random1_test (word nseed)
{
  long a;
  word d, w;


  nfail = 0;
  cprintf("random %d...", nseed);
  seed = nseed;
  for (a=0; a<RAMSIZE; a+=2)
    {
      w = random();
      wb(a, w);
    }
  seed = nseed;
  for (a=0; a<RAMSIZE; a+=2)
    {
      w = random();
      RBE(a, d, w)
    }
  report();
}

static void
random_test ()
{
  random1_test (23);
  random1_test (57);
  random1_test (89);
}

refresh_test ()
{
  long a;
  word d, w;

  nfail = 0;
  cprintf("refresh...", w);
  for (a=0; a<RAMSIZE; a+=26)
    {
      w = (word)a | (word)(a>>16);
      wb(a, w);
    }
  p10.b = 0;
  cprintf("pause...");
  for (d=0; d<20; d++)
    {
      p10.b = d & 1;
      wait_ms(500);
    }
  for (a=0; a<RAMSIZE; a+=26)
    {
      w = (word)a | (word)(a>>16);
      RBE(a, d, w)
    }
  report();
}

alignment_test ()
{
  long a;
  word d, w;

  nfail = 0;
  cprintf("alignment...", w);
  for (a=0; a<RAMSIZE; a+=2)
    {
      w = (word)a + 1;
      wb(a + 1, w);
    }
  for (a=0; a<RAMSIZE; a+=2)
    {
      w = (word)a + 1;
      RBE(a+1, d, w)
    }
  report();
}

alignwrite_test ()
{
  long a;
  word d, w;
  word *r1, *r2;

  nfail = 0;
  cprintf("alignment...", w);
  for (a=0; a<RAMSIZE-6; a+=2)
    {
      int w2;
      void *p = long2ptr (a + RAM_ADDR);
      w = (word)a + 1;
      w2 = ~w;
      asm("\
	mov.w %1,1[%3] \n\
 	mov.w %1,5[%3] \n\
 	mov.w %2,3[%3] \n\
 	mov.w 3[%3],%0 \n\
	mov.w #2,0x104006" : "=Rhi" (d) : "Rhi" (w2), "Rhi" (w), "Ra1" (p));
      la->status = 0;
      if (d != w) {
	word d2 = rb(a+3);
	FAIL (a, d, d2, w);
	la_dump();
      }

    }
  report();
}

main()
{
  unsigned int i, j, t, u, done;
  long count, fails;
  int h, m, l;
  unsigned int oi, oj, op=-1;
  unsigned char new_pd9;
  long addr = 0;
  volatile word *r1, *r2;

  p10.b = 0x33;

  F_IO_0 = 1; /* This is RESET ACTIVE HIGH */
  F_IO_1 = 1; /* chip select ACTIVE LOW */
  F_IO_2 = 0; /* increment phase */
  F_IO_3 = 0; /* decrement phase */

  prcr.b = 15;

  pm0.b = 0x05;
  pm1.b = 0x00;
  cm0   = 0x00;
  ds = 0x0f;
  mcd = 0x12; /* divide by 1 */

  ewcr0 = 0x01; /* zero wait states */
  ewcr1 = 0x01;
  ewcr2 = 0x01;
  ewcr3 = 0x01;

  new_pd9 = pd9.b | 0x0f;
  prcr.b = 0x04;
  pd9.b = new_pd9; /* f_io_* */

  ndumps = 0;
  wait_ms (1);

#if 1
  while (tty_ready ())
    tty_readc ();
  cprintf("about to start...\n");
  tty_readc ();
#endif

  F_IO_0 = 0; /* de-assert reset */
  F_IO_1 = 0; /* chip select ACTIVE LOW */

  cprintf("running...\n");

  RUN();

  totalfail = 0;
  count = 0;
  while (! tty_ready ())
    {
#define C if (tty_ready()) break;
      nfail = 0;
      count ++;
      cprintf("\033[34mPass %d\033[0m\n", count);

      C zeros_test();
      C ones_test();
      C aa55_test();
      C msb_test();
      C lsb_test();
      C random_test();
      C refresh_test();
      C alignment_test();
      C alignwrite_test();
#if 0
      C walking_ones_test();
      C walking_zeros_test();
#endif
    }

  prcr.b = 15;
  pm0.pm0mode = 0; /* single chip */
  prcr.b = 0;

}
 
